type program = exp
and exp = 
  | CONST of int
  | VAR of var
  | ADD of exp * exp
  | SUB of exp * exp
  | READ
  | ISZERO of exp
  | IF of exp * exp * exp
  | LET of var * exp * exp
and var = string

(* 1 *)
let pgm1 = CONST 1 

(* (5+2)+(3-10) *)
let pgm2 = ADD (ADD (CONST 5, CONST 2), SUB (CONST 3, CONST 10))

(* if iszero 1 then 3 else 4 *)
let pgm3 = IF (ISZERO (CONST 1), CONST 3, CONST 4) 

(* let x = 5 
   in x - 3 *)
let pgm4 = LET ("x", CONST 5, SUB(VAR "x", CONST 3)) 

(* let z = 5
   in let x = 3
      in let y = x - 1  // 2
         in let x = 4
            in z - (x - y)  // 3
*)
let pgm5 = 
  LET ("z", CONST 5, 
    LET ("x", CONST 3, 
      LET ("y", SUB (VAR "x", CONST 1), 
        LET ("x", CONST 4, 
          SUB (VAR "z", (SUB (VAR "x", VAR "y")))))))

(*
let x = 7 
in let y = 2
   in let y = (let x = x - 1 in x - y)
      in  (x - 8 - y)  // -5
*)
let pgm6 = 
  LET ("x", CONST 7, 
    LET ("y", CONST 2, 
      LET ("y", LET ("x", SUB(VAR "x", CONST 1), 
                  SUB (VAR "x", VAR "y")), 
        SUB (SUB (VAR "x", CONST 8), VAR "y"))))

(*
  let x = read
  in let y = 2 
     in x + y
*)
let pgm7 =
  LET ("x", READ,
    LET ("y", CONST 2,
      ADD (VAR "x", VAR "y")))

(* iszero (iszero 1)  *)
let pgm8 = ISZERO (ISZERO (CONST 1))

(* if 1 then 2 else 3 *)
let pgm9 = IF (CONST 1, CONST 2, CONST 3)

(* let x = 1 in x + y *)
let pgm10 = LET ("x", CONST 1, ADD (VAR "x", VAR "y"))

type value = Int of int | Bool of bool
let value2str v = match v with Int n -> string_of_int n | Bool b -> string_of_bool b

type env = (var * value) list
let empty_env = []
let extend_env x v e = (x,v)::e
let rec lookup_env x e = 
  match e with
  | [] -> raise (Failure ("variable " ^ x ^ " is not bound in env")) 
  | (y,v)::tl -> if x = y then v else lookup_env x tl

let rec eval_bop : (int -> int -> int) -> exp -> exp -> env -> value
=fun op e1 e2 env -> 
  let v1 = eval e1 env in
  let v2 = eval e2 env in
    (match v1,v2 with
    | Int n1, Int n2 -> Int (op n1 n2)
    | _ -> raise (Failure "Type Error: non-numeric values"))

and eval : exp -> env -> value
=fun exp env ->
  match exp with
  | CONST n -> Int n
  | VAR x -> lookup_env x env
  | ADD (e1,e2) -> eval_bop (+) e1 e2 env
  | SUB (e1,e2) -> eval_bop (-) e1 e2 env
  | READ -> Int (read_int())
  | ISZERO e -> 
    (match eval e env with
    | Int n -> if n = 0 then Bool true else Bool false
    | _ -> raise (Failure "Type error: iszero must take bool"))
  | IF (e1,e2,e3) -> 
    (match eval e1 env with
    | Bool true -> eval e2 env
    | Bool false -> eval e3 env
    | _ -> raise (Failure "Type Error: condition must be Bool type"))
  | LET (x,e1,e2) ->
    let v1 = eval e1 env in
    let env' = extend_env x v1 env in
      eval e2 env'

let run : program -> value
=fun pgm -> eval pgm empty_env
